/******************************************************************/
/* This program is affiliated with the second tutorial on pixels  */
/* This program plots pixels four (4) different ways in C and ASM */
/*								  */
/* tcc -2 tut2.c                                                  */
/******************************************************************/

/* The only header we need to include right now */

#include <dos.h>
#include <mem.h>
#include <conio.h>

/* Define our constants */

#define VGA     0x0013
#define TXT     0x0003
#define TRUE    1
#define FALSE   0
#define WIDTH   320
#define HEIGHT  200
#define SEGMENT 0xA000

/* Our typedef and pointer to far video memory */

typedef unsigned char BYTE;
BYTE far *video_buffer = (unsigned char*)0xA0000000L;

/* Our function prototypes */

void Plot_Pixel_C_Slow(int x, int y, BYTE color);
void Plot_Pixel_C_Fast(int x, int y, BYTE color);
void Plot_Pixel_Asm_Slow(int x, int y, BYTE color);
void Plot_Pixel_Asm_Fast(int x, int y, BYTE color);
void Fill_Screen_C_Slow(BYTE color);
void Fill_Screen_C_Fast(BYTE color);
void Fill_Screen_Asm_Fast(BYTE color);
void Set_Mode(int mode);

/* main() */

void main() {
  Set_Mode(VGA);			/* set Mode13h */

  Plot_Pixel_C_Slow(1, 1, 2);		/* plot pixels in 4 ways */
  Plot_Pixel_C_Fast(3, 3, 3);
  Plot_Pixel_Asm_Slow(5, 5, 4);
  Plot_Pixel_Asm_Fast(7, 7, 5);

  while(!kbhit());

  Set_Mode(TXT);			/* return to Mode03h */
}

/**************************************************************/
/* This is the SLOW version of plotting a pixel in C          */
/**************************************************************/

void Plot_Pixel_C_Slow(int x, int y, BYTE color) {
  video_buffer[(y * 320) + x] = color;
}

/**************************************************************/
/* This is a FASTER version of plotting a pixel in C          */
/**************************************************************/

void Plot_Pixel_C_Fast(int x, int y, BYTE color) {
  video_buffer[(y << 8) + ( y << 6) + x] = color;
}

/**************************************************************/
/* This is probably the slowest function in Assembly          */
/**************************************************************/

void Plot_Pixel_Asm_Slow(int x, int y, BYTE color) {
  asm {
    mov ah, 0ch
    mov al, color
    mov bh, 0
    mov cx, x
    mov dx, y
    int 10h
  }
}

/**************************************************************/
/* This is a very fast function in Assembly = direct memory   */
/**************************************************************/

void Plot_Pixel_Asm_Fast(int x, int y, BYTE color) {
  asm {
    mov ax, 0xA000              // load the screen segment into ax because
    mov es, ax                  // you can't load Segment regs directly 
    mov di, [y]                 // now, we'll start to calculate the offset
    mov bx, di                  // reg / reg is faster than mem / reg
    shl di, 6                   // shift di six bits to the left
    shl bx, 8                   // shift bx eight bits to the left
    add di, bx                  // di = y * 320
    add di, [x]                 // add x to the offset
    mov al, [color]             // put color into AL
    mov es:[di], al             // plot straight to memory
  }
}

/**************************************************************/
/* This is a slow way to fill a screen in C                   */
/**************************************************************/

void Fill_Screen_C_Slow(BYTE color) {
  int x, y;

  for(x=0;x<WIDTH;x++) {
    for(y=0;y<HEIGHT;y++) {
      Plot_Pixel_Asm_Fast(x, y, color);
    }
  }
}

/**************************************************************/
/* This is a fast way to fill a screen in C                   */
/**************************************************************/

void Fill_Screen_C_Fast(BYTE color) {
  _fmemset(video_buffer, color, WIDTH*HEIGHT+1);
}

/**************************************************************/
/* This is a fast way to fill the screen in Assembly          */
/**************************************************************/

void Fill_Screen_Asm_Fast(BYTE color) {
  asm {
    les di, [SEGMENT]
    mov al, BYTE PTR color
    mov ah, al
    mov cx, 32000
    rep stosw
  }
}

/**************************************************************/
/* This is the function to access Mode13h and Mode03h         */
/**************************************************************/

void Set_Mode(int mode) {
  asm {
    mov ax, mode
    int 10h
  }
}